#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_product."""

import abc
import datetime
import sys
import uuid
import warnings

import matplotlib.pyplot as plt

import networkx as nx

import plotly.figure_factory as ff
import plotly.graph_objects as go

from .base_component import BaseComponent, BaseComponentState


class BaseProduct(object, metaclass=abc.ABCMeta):
    """BaseProduct.

    BaseProduct class for expressing target product in a project.
    BaseProduct consists of multiple BaseComponents.
    This class will be used as a template.

    Args:
        name (str, optional): Name of this product. Defaults to None -> "Product".
        ID (str, optional): ID will be defined automatically. Defaults to None -> str(uuid.uuid4()).
        component_set (set[BaseComponent], optional): Set of BaseComponents.
    """

    def __init__(
        self, name: str = None, ID: str = None, component_set: set[BaseComponent] = None
    ):
        """init."""
        # ----
        # Constraint parameters on simulation
        # --
        # Basic parameter
        self.name = name if name is not None else "Product"
        self.ID = ID if ID is not None else str(uuid.uuid4())

        self.component_set = set()
        if component_set is not None:
            self.update_component_set(component_set)

    def initialize(self, state_info: bool = True, log_info: bool = True):
        """
        Initialize the changeable variables of BaseProduct.

        BaseComponent in `component_set` are also initialized by this function.

        Args:
            state_info (bool, optional): Whether to initialize state information. Defaults to True.
            log_info (bool, optional): Whether to initialize log information. Defaults to True.
        """
        for c in self.component_set:
            c.initialize(state_info=state_info, log_info=log_info)

    def __str__(self):
        """Return the name list of BaseComponent.

        Returns:
            str: Name list of BaseComponent.
        """
        return f"{[str(c) for c in self.component_set]}"

    def append_component(self, component: BaseComponent):
        """
        Append target component to this workflow.

        .. deprecated:: Use add_component instead.

        Args:
            component (BaseComponent): Target component.
        """
        warnings.warn(
            "append_component is deprecated, use add_component instead.",
            DeprecationWarning,
        )
        self.add_component(component)

    def extend_component_list(self, component_set: set[BaseComponent]):
        """
        Extend target component_set to this product.

        .. deprecated:: Use update_component_set instead.

        Args:
            component_set (set[BaseComponent]): Target component set.
        """
        warnings.warn(
            "extend_component_list is deprecated, use update_component_set instead.",
            DeprecationWarning,
        )
        for component in component_set:
            self.add_component(component)

    def add_component(self, component: BaseComponent):
        """
        Add target component to this workflow.

        Args:
            component (BaseComponent): Target component.
        """
        if not isinstance(component, BaseComponent):
            raise TypeError(
                f"component must be BaseComponent, but got {type(component)}"
            )
        self.component_set.add(component)
        component.parent_product_id = self.ID

    def update_component_set(self, component_set: set[BaseComponent]):
        """
        Update target component_set to this product.

        Args:
            component_set (set[BaseComponent]): Target component set.
        """
        for component in component_set:
            if not isinstance(component, BaseComponent):
                raise TypeError(
                    f"All elements of component_set must be BaseComponent, but found {type(component)}"
                )
            self.component_set.add(component)
            component.parent_product_id = self.ID

    def create_component(
        self,
        # Basic parameters
        name: str = None,
        ID: str = None,
        child_component_id_set: set[str] = None,
        targeted_task_id_set: set[str] = None,
        space_size: float = None,
        # Basic variables
        state: BaseComponentState = BaseComponentState.NONE,
        state_record_list: list[BaseComponentState] = None,
        placed_workplace_id: str = None,
        placed_workplace_id_record_list: list[str] = None,
        # Advanced parameters for customized simulation
        error_tolerance: float = None,
        # Advanced variables for customized simulation
        error: float = None,
    ):
        """
        Create BaseComponent instance and add it to this product.

        Args:
            name (str, optional): Name of this component. Defaults to None -> "New Component".
            ID (str, optional): ID will be defined automatically.
            child_component_id_set (set[str], optional): Child BaseComponents id set. Defaults to None -> set().
            targeted_task_id_set (set[str], optional): Targeted tasks id set. Defaults to None -> set().
            space_size (float, optional): Space size related to base_workplace's max_space_size. Default to None -> 1.0.
            state (BaseComponentState, optional): State of this task in simulation. Defaults to BaseComponentState.NONE.
            state_record_list (List[BaseComponentState], optional): Record list of state. Defaults to None -> [].
            placed_workplace_id (str, optional): A workplace which this component is placed in simulation. Defaults to None.
            placed_workplace_id_record_list (List[str], optional): Record of placed workplace ID in simulation. Defaults to None -> [].
            error_tolerance (float, optional): Advanced parameter.
            error (float, optional): Advanced variables.

        Returns:
            BaseComponent: The created component.
        """
        component = BaseComponent(
            name=name,
            ID=ID,
            child_component_id_set=child_component_id_set,
            targeted_task_id_set=targeted_task_id_set,
            space_size=space_size,
            state=state,
            state_record_list=state_record_list,
            placed_workplace_id=placed_workplace_id,
            placed_workplace_id_record_list=placed_workplace_id_record_list,
            error_tolerance=error_tolerance,
            error=error,
        )
        self.add_component(component)
        return component

    def export_dict_json_data(self):
        """
        Export the information of this product to JSON data.

        Returns:
            dict: JSON format data.
        """
        dict_json_data = {}
        dict_json_data.update(
            type=self.__class__.__name__,
            name=self.name,
            ID=self.ID,
            component_set=[c.export_dict_json_data() for c in self.component_set],
        )
        return dict_json_data

    def read_json_data(self, json_data: dict):
        """
        Read the JSON data for creating BaseProduct instance.

        Args:
            json_data (dict): JSON data.
        """
        self.name = json_data["name"]
        self.ID = json_data["ID"]
        j_list = json_data["component_set"]
        self.component_set = {
            BaseComponent(
                name=j["name"],
                ID=j["ID"],
                child_component_id_set=set(j["child_component_id_set"]),
                targeted_task_id_set=set(j["targeted_task_id_set"]),
                space_size=j["space_size"],
                state=BaseComponentState(j["state"]),
                state_record_list=[
                    BaseComponentState(num) for num in j["state_record_list"]
                ],
                placed_workplace_id=j["placed_workplace_id"],
                placed_workplace_id_record_list=j["placed_workplace_id_record_list"],
            )
            for j in j_list
        }

    def extract_none_component_set(self, target_time_list: list[int]):
        """
        Extract NONE component set from simulation result.

        Args:
            target_time_list (List[int]): Target time list. If you want to extract none component from time 2 to time 4, you must set [2, 3, 4] to this argument.

        Returns:
            List[BaseComponent]: List of BaseComponent.
        """
        return self.__extract_state_component_set(
            target_time_list, BaseComponentState.NONE
        )

    def extract_ready_component_set(self, target_time_list: list[int]):
        """
        Extract READY component set from simulation result.

        Args:
            target_time_list (List[int]): Target time list. If you want to extract ready component from time 2 to time 4, you must set [2, 3, 4] to this argument.

        Returns:
            List[BaseComponent]: List of BaseComponent.
        """
        return self.__extract_state_component_set(
            target_time_list, BaseComponentState.READY
        )

    def extract_working_component_set(self, target_time_list: list[int]):
        """
        Extract WORKING component set from simulation result.

        Args:
            target_time_list (List[int]): Target time list. If you want to extract working component from time 2 to time 4, you must set [2, 3, 4] to this argument.

        Returns:
            List[BaseComponent]: List of BaseComponent.
        """
        return self.__extract_state_component_set(
            target_time_list, BaseComponentState.WORKING
        )

    def extract_finished_component_set(self, target_time_list: list[int]):
        """
        Extract FINISHED component set from simulation result.

        Args:
            target_time_list (List[int]): Target time list. If you want to extract finished component from time 2 to time 4, you must set [2, 3, 4] to this argument.

        Returns:
            List[BaseComponent]: List of BaseComponent.
        """
        return self.__extract_state_component_set(
            target_time_list, BaseComponentState.FINISHED
        )

    def __extract_state_component_set(
        self, target_time_list: list[int], target_state: BaseComponentState
    ):
        """
        Extract state component set from simulation result.

        Args:
            target_time_list (List[int]): Target time list. If you want to extract target_state component from time 2 to time 4, you must set [2, 3, 4] to this argument.
            target_state (BaseComponentState): Target state.

        Returns:
            List[BaseComponent]: List of BaseComponent.
        """
        return {
            component
            for component in self.component_set
            if all(
                len(component.state_record_list) > time
                and component.state_record_list[time] == target_state
                for time in target_time_list
            )
        }

    def reverse_log_information(self):
        """Reverse log information of all."""
        for c in self.component_set:
            c.reverse_log_information()

    def record(self, working=True):
        """Record placed workplace id in this time.

        Args:
            working (bool, optional): Whether to record as working. Defaults to True.
        """
        for c in self.component_set:
            c.record_placed_workplace_id()
            c.record_state(working=working)

    def remove_absence_time_list(self, absence_time_list: list[int]):
        """
        Remove record information on `absence_time_list`.

        Args:
            absence_time_list (List[int]): List of absence step time in simulation.
        """
        for c in self.component_set:
            c.remove_absence_time_list(absence_time_list)

    def insert_absence_time_list(self, absence_time_list: list[int]):
        """
        Insert record information on `absence_time_list`.

        Args:
            absence_time_list (List[int]): List of absence step time in simulation.
        """
        for c in self.component_set:
            c.insert_absence_time_list(absence_time_list)

    def print_log(self, target_step_time: int):
        """
        Print log in `target_step_time`.

        Args:
            target_step_time (int): Target step time of printing log.
        """
        for component in self.component_set:
            component.print_log(target_step_time)

    def print_all_log_in_chronological_order(self, backward: bool = False):
        """
        Print all log in chronological order.

        Args:
            backward (bool, optional): If True, print logs in reverse order. Defaults to False.
        """
        if len(self.component_set) > 0:
            sample_component = next(iter(self.component_set))
            n = len(sample_component.state_record_list)
            if backward:
                for i in range(n):
                    t = n - 1 - i
                    print("TIME: ", t)
                    self.print_log(t)
            else:
                for t in range(n):
                    print("TIME: ", t)
                    self.print_log(t)

    def plot_simple_gantt(
        self,
        target_id_order_list: list[str] = None,
        finish_margin: float = 1.0,
        print_product_name: bool = True,
        view_ready: bool = True,
        component_color: str = "#FF6600",
        ready_color: str = "#C0C0C0",
        figsize: tuple[float, float] = None,
        dpi: float = 100.0,
        save_fig_path: str = None,
    ):
        """
        Plot Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            target_id_order_list (list[str], optional): Target ID order list. Defaults to None.
            finish_margin (float, optional): Margin of finish time in Gantt chart. Defaults to 1.0.
            print_product_name (bool, optional): Print product name or not. Defaults to True.
            view_ready (bool, optional): View READY time or not. Defaults to True.
            component_color (str, optional): Component color setting information. Defaults to "#FF6600".
            ready_color (str, optional): Ready color setting information. Defaults to "#C0C0C0".
            figsize ((float, float), optional): Width, height in inches. Default to None -> [6.4, 4.8].
            dpi (float, optional): The resolution of the figure in dots-per-inch. Default to 100.0.
            save_fig_path (str, optional): Path of saving figure. Defaults to None.

        Returns:
            fig: Figure in plt.subplots().
        """
        if figsize is None:
            figsize = [6.4, 4.8]
        fig, gnt = self.create_simple_gantt(
            target_id_order_list=target_id_order_list,
            finish_margin=finish_margin,
            print_product_name=print_product_name,
            view_ready=view_ready,
            component_color=component_color,
            ready_color=ready_color,
            figsize=figsize,
            dpi=dpi,
            save_fig_path=save_fig_path,
        )
        _ = gnt
        return fig

    def create_simple_gantt(
        self,
        target_id_order_list: list[str] = None,
        finish_margin: float = 1.0,
        print_product_name: bool = True,
        view_ready: bool = True,
        component_color: str = "#FF6600",
        ready_color: str = "#C0C0C0",
        figsize: tuple[float, float] = None,
        dpi: float = 100.0,
        save_fig_path: str = None,
    ):
        """
        Create Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            target_id_order_list (list[str], optional): Target ID order list. Defaults to None.
            finish_margin (float, optional): Margin of finish time in Gantt chart. Defaults to 1.0.
            view_ready (bool, optional): View READY time or not. Defaults to True.
            print_product_name (bool, optional): Print product name or not. Defaults to True.
            component_color (str, optional): Component color setting information. Defaults to "#FF6600".
            ready_color (str, optional): Ready color setting information. Defaults to "#C0C0C0".
            figsize ((float, float), optional): Width, height in inches. Default to None -> [6.4, 4.8].
            dpi (float, optional): The resolution of the figure in dots-per-inch. Default to 100.0.
            save_fig_path (str, optional): Path of saving figure. Defaults to None.

        Returns:
            fig: Figure in plt.subplots().
            gnt: Axes in plt.subplots().
        """
        if figsize is None:
            figsize = [6.4, 4.8]
        fig, gnt = plt.subplots()
        fig.figsize = figsize
        fig.dpi = dpi
        gnt.set_xlabel("step")
        gnt.grid(True)

        target_instance_list = self.component_set
        if target_id_order_list is not None:
            id_to_instance = {instance.ID: instance for instance in self.component_set}
            target_instance_list = [
                id_to_instance[tid]
                for tid in target_id_order_list
                if tid in id_to_instance
            ]

        target_instance_list = list(reversed(list(target_instance_list)))

        y_ticks = [10 * (n + 1) for n in range(len(target_instance_list))]
        y_tick_labels = [com.name for com in target_instance_list]
        if print_product_name:
            y_tick_labels = [f"{self.name}: {com.name}" for com in target_instance_list]

        gnt.set_yticks(y_ticks)
        gnt.set_yticklabels(y_tick_labels)

        for time, c in enumerate(target_instance_list):
            (
                ready_time_list,
                working_time_list,
            ) = c.get_time_list_for_gantt_chart(finish_margin=finish_margin)
            if view_ready:
                gnt.broken_barh(
                    ready_time_list, (y_ticks[time] - 5, 9), facecolors=(ready_color)
                )
            gnt.broken_barh(
                working_time_list, (y_ticks[time] - 5, 9), facecolors=(component_color)
            )

        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig, gnt

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        target_id_order_list: list[str] = None,
        finish_margin: float = 1.0,
        print_product_name: bool = True,
        view_ready: bool = False,
    ):
        """
        Create data for gantt plotly from component_set.

        Args:
            init_datetime (datetime.datetime): Start datetime of project.
            unit_timedelta (datetime.timedelta): Unit time of simulation.
            target_id_order_list (list[str], optional): Target ID order list. Defaults to None.
            finish_margin (float, optional): Margin of finish time in Gantt chart. Defaults to 1.0.
            print_product_name (bool, optional): Print product name or not. Defaults to True.
            view_ready (bool, optional): View READY time or not. Defaults to False.

        Returns:
            List[dict]: Gantt plotly information of this BaseProduct.
        """
        df = []
        target_instance_list = self.component_set
        if target_id_order_list is not None:
            id_to_instance = {instance.ID: instance for instance in self.component_set}
            target_instance_list = [
                id_to_instance[tid]
                for tid in target_id_order_list
                if tid in id_to_instance
            ]
        for component in target_instance_list:
            (
                ready_time_list,
                working_time_list,
            ) = component.get_time_list_for_gantt_chart(finish_margin=finish_margin)

            task_name = component.name
            if print_product_name:
                task_name = f"{self.name}: {component.name}"

            if view_ready:
                for from_time, length in ready_time_list:
                    to_time = from_time + length
                    df.append(
                        {
                            "Task": task_name,
                            "Start": (
                                init_datetime + from_time * unit_timedelta
                            ).strftime("%Y-%m-%d %H:%M:%S"),
                            "Finish": (
                                init_datetime + to_time * unit_timedelta
                            ).strftime("%Y-%m-%d %H:%M:%S"),
                            "State": "READY",
                            "Type": "Component",
                        }
                    )
            for from_time, length in working_time_list:
                to_time = from_time + length
                df.append(
                    {
                        "Task": task_name,
                        "Start": (init_datetime + from_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "Finish": (init_datetime + to_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "State": "WORKING",
                        "Type": "Component",
                    }
                )
        return df

    def create_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        target_id_order_list: list[str] = None,
        title: str = "Gantt Chart",
        colors: dict[str, str] = None,
        index_col: str = None,
        showgrid_x: bool = True,
        showgrid_y: bool = True,
        group_tasks: bool = True,
        show_colorbar: bool = True,
        finish_margin: float = 1.0,
        print_product_name: bool = True,
        view_ready: bool = False,
        save_fig_path: str = None,
    ):
        """
        Create Gantt chart by plotly.

        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime): Start datetime of project.
            unit_timedelta (datetime.timedelta): Unit time of simulation.
            target_id_order_list (list[str], optional): Target ID order list. Defaults to None.
            title (str, optional): Title of Gantt chart. Defaults to "Gantt Chart".
            colors (Dict[str, str], optional): Color setting of plotly Gantt chart. Defaults to None -> dict(Component="rgb(246, 37, 105)", READY="rgb(107, 127, 135)").
            index_col (str, optional): index_col of plotly Gantt chart. Defaults to None -> "State".
            showgrid_x (bool, optional): showgrid_x of plotly Gantt chart. Defaults to True.
            showgrid_y (bool, optional): showgrid_y of plotly Gantt chart. Defaults to True.
            group_tasks (bool, optional): group_tasks of plotly Gantt chart. Defaults to True.
            show_colorbar (bool, optional): show_colorbar of plotly Gantt chart. Defaults to True.
            finish_margin (float, optional): Margin of finish time in Gantt chart. Defaults to 1.0.
            print_product_name (bool, optional): Print product name or not. Defaults to True.
            view_ready (bool, optional): View READY time or not. Defaults to False.
            save_fig_path (str, optional): Path of saving figure. Defaults to None.

        Returns:
            figure: Figure for a gantt chart.
        """
        colors = (
            colors
            if colors is not None
            else {"WORKING": "rgb(246, 37, 105)", "READY": "rgb(107, 127, 135)"}
        )
        index_col = index_col if index_col is not None else "State"
        df = self.create_data_for_gantt_plotly(
            init_datetime,
            unit_timedelta,
            target_id_order_list=target_id_order_list,
            finish_margin=finish_margin,
            print_product_name=print_product_name,
            view_ready=view_ready,
        )
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            dot_point = save_fig_path.rfind(".")
            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            else:
                fig.write_image(save_fig_path)

        return fig

    def get_networkx_graph(self):
        """
        Get the information of networkx graph.

        Returns:
            networkx.DiGraph: Directed graph of the product.
        """
        g = nx.DiGraph()

        # 1. add all nodes
        for component in self.component_set:
            g.add_node(component)

        id_to_component = {c.ID: c for c in self.component_set}

        # 2. add all edges
        for component in self.component_set:
            for child_c_id in component.child_component_id_set:
                child_c = id_to_component.get(child_c_id, None)
                if child_c is not None:
                    g.add_edge(component, child_c)

        return g

    def draw_networkx(
        self,
        g: nx.DiGraph = None,
        pos: dict = None,
        arrows: bool = True,
        component_node_color: str = "#FF6600",
        figsize: tuple[float, float] = None,
        dpi: float = 100.0,
        save_fig_path: str = None,
        **kwargs,
    ):
        """
        Draw networkx.

        Args:
            g (networkx.Digraph, optional): The information of networkx graph. Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional): Layout of networkx. Defaults to None -> networkx.spring_layout(g).
            arrows (bool, optional): Digraph or Graph(no arrows). Defaults to True.
            component_node_color (str, optional): Node color setting information. Defaults to "#FF6600".
            figsize ((float, float), optional): Width, height in inches. Default to None -> [6.4, 4.8].
            dpi (float, optional): The resolution of the figure in dots-per-inch. Default to 100.0.
            save_fig_path (str, optional): Path of saving figure. Defaults to None.
            **kwargs: Other networkx settings.

        Returns:
            figure: Figure for a network.
        """
        if figsize is None:
            figsize = [6.4, 4.8]
        fig = plt.figure(figsize=figsize, dpi=dpi)
        g = g if g is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(g)

        # component
        component_set = self.component_set
        nx.draw_networkx_nodes(
            g,
            pos,
            nodelist=component_set,
            node_color=component_node_color,
            **kwargs,
        )

        nx.draw_networkx_labels(g, pos, **kwargs)
        nx.draw_networkx_edges(g, pos, arrows=arrows, **kwargs)
        plt.axis("off")
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig

    def get_node_and_edge_trace_for_plotly_network(
        self,
        g: nx.DiGraph = None,
        pos: dict = None,
        node_size: int = 20,
        component_node_color: str = "#FF6600",
    ):
        """
        Get nodes and edges information of plotly network.

        Args:
            g (networkx.Digraph, optional): The information of networkx graph. Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional): Layout of networkx. Defaults to None -> networkx.spring_layout(g).
            node_size (int, optional): Node size setting information. Defaults to 20.
            component_node_color (str, optional): Node color setting information. Defaults to "#FF6600".

        Returns:
            node_trace: Node information of plotly network.
            edge_trace: Edge information of plotly network.
        """
        g = g if g is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(g)

        node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            marker={"color": component_node_color, "size": node_size},
        )

        for node in g.nodes:
            x, y = pos[node]
            node_trace["x"] = node_trace["x"] + (x,)
            node_trace["y"] = node_trace["y"] + (y,)
            node_trace["text"] = node_trace["text"] + (node,)

        edge_trace = go.Scatter(
            x=[],
            y=[],
            line={"width": 1, "color": "#888"},
            mode="lines",
        )

        for edge in g.edges:
            x = edge[0]
            y = edge[1]
            x_pos_x, x_pos_y = pos[x]
            y_pos_x, y_pos_y = pos[y]
            edge_trace["x"] += (x_pos_x, y_pos_x)
            edge_trace["y"] += (x_pos_y, y_pos_y)
        return node_trace, edge_trace

    def draw_plotly_network(
        self,
        g: nx.DiGraph = None,
        pos: dict = None,
        title: str = "Product",
        node_size: int = 20,
        component_node_color: str = "#FF6600",
        save_fig_path: str = None,
    ):
        """
        Draw plotly network.

        Args:
            g (networkx.Digraph, optional): The information of networkx graph. Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional): Layout of networkx. Defaults to None -> networkx.spring_layout(g).
            title (str, optional): Figure title of this network. Defaults to "Product".
            node_size (int, optional): Node size setting information. Defaults to 20.
            component_node_color (str, optional): Node color setting information. Defaults to "#FF6600".
            save_fig_path (str, optional): Path of saving figure. Defaults to None.

        Returns:
            figure: Figure for a network.
        """
        g = g if g is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(g)
        node_trace, edge_trace = self.get_node_and_edge_trace_for_plotly_network(
            g, pos, node_size=node_size, component_node_color=component_node_color
        )
        fig = go.Figure(
            data=[edge_trace, node_trace],
            layout=go.Layout(
                title=title,
                showlegend=False,
                annotations=[
                    {
                        "ax": edge_trace["x"][index * 2],
                        "ay": edge_trace["y"][index * 2],
                        "axref": "x",
                        "ayref": "y",
                        "x": edge_trace["x"][index * 2 + 1],
                        "y": edge_trace["y"][index * 2 + 1],
                        "xref": "x",
                        "yref": "y",
                        "showarrow": True,
                        "arrowhead": 5,
                    }
                    for index in range(0, int(len(edge_trace["x"]) / 2))
                ],
                xaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
                yaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
            ),
        )
        if save_fig_path is not None:
            dot_point = save_fig_path.rfind(".")
            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            else:
                fig.write_image(save_fig_path)

        return fig

    def get_target_component_mermaid_diagram(
        self,
        target_component_set: set[BaseComponent],
        shape_component: str = "odd",
        link_type_str: str = "-->",
        subgraph: bool = True,
        subgraph_direction: str = "LR",
    ):
        """
        Get mermaid diagram of target component.

        Args:
            target_component_set (set[BaseComponent]): Target component set.
            shape_component (str, optional): Shape of mermaid diagram. Defaults to "odd".
            link_type_str (str, optional): Link type of mermaid diagram. Defaults to "-->".
            subgraph (bool, optional): Subgraph or not. Defaults to True.
            subgraph_direction (str, optional): Direction of subgraph. Defaults to "LR".

        Returns:
            list[str]: List of lines for mermaid diagram.
        """

        list_of_lines = []
        if subgraph:
            list_of_lines.append(f"subgraph {self.ID}[{self.name}]")
            list_of_lines.append(f"direction {subgraph_direction}")

        for component in target_component_set:
            if component in self.component_set:
                list_of_lines.extend(
                    component.get_mermaid_diagram(
                        shape=shape_component,
                    )
                )

        for component in target_component_set:
            if component in self.component_set:
                for child_component_id in component.child_component_id_set:
                    if child_component_id in [c.ID for c in target_component_set]:
                        list_of_lines.append(
                            f"{component.ID}{link_type_str}{child_component_id}"
                        )

        if subgraph:
            list_of_lines.append("end")

        return list_of_lines

    def get_mermaid_diagram(
        self,
        shape_component: str = "odd",
        link_type_str: str = "-->",
        subgraph: bool = True,
        subgraph_direction: str = "LR",
    ):
        """
        Get mermaid diagram of this product.

        Args:
            shape_component (str, optional): Shape of mermaid diagram. Defaults to "odd".
            link_type_str (str, optional): Link type of mermaid diagram. Defaults to "-->".
            subgraph (bool, optional): Subgraph or not. Defaults to True.
            subgraph_direction (str, optional): Direction of subgraph. Defaults to "LR".

        Returns:
            list[str]: List of lines for mermaid diagram.
        """

        list_of_lines = []
        if subgraph:
            list_of_lines.append(f"subgraph {self.ID}[{self.name}]")
            list_of_lines.append(f"direction {subgraph_direction}")

        for component in self.component_set:
            list_of_lines.extend(
                component.get_mermaid_diagram(
                    shape=shape_component,
                )
            )

        for component in self.component_set:
            for child_component_id in component.child_component_id_set:
                list_of_lines.append(
                    f"{component.ID}{link_type_str}{child_component_id}"
                )

        if subgraph:
            list_of_lines.append("end")

        return list_of_lines

    def print_target_component_mermaid_diagram(
        self,
        target_component_set: set[BaseComponent],
        orientations: str = "LR",
        shape_component: str = "odd",
        link_type_str: str = "-->",
        subgraph: bool = True,
        subgraph_direction: str = "LR",
    ):
        """
        Print mermaid diagram of target component.

        Args:
            target_component_set (set[BaseComponent]): Target component set.
            orientations (str, optional): Orientation of mermaid diagram. See: https://mermaid.js.org/syntax/flowchart.html#direction. Defaults to "LR".
            shape_component (str, optional): Shape of mermaid diagram. Defaults to "odd".
            link_type_str (str, optional): Link type of mermaid diagram. Defaults to "-->".
            subgraph (bool, optional): Subgraph or not. Defaults to True.
            subgraph_direction (str, optional): Direction of subgraph. Defaults to "LR".
        """
        print(f"flowchart {orientations}")
        list_of_lines = self.get_target_component_mermaid_diagram(
            target_component_set=target_component_set,
            shape_component=shape_component,
            link_type_str=link_type_str,
            subgraph=subgraph,
            subgraph_direction=subgraph_direction,
        )
        print(*list_of_lines, sep="\n")

    def print_mermaid_diagram(
        self,
        orientations: str = "LR",
        shape_component: str = "odd",
        link_type_str: str = "-->",
        subgraph: bool = True,
        subgraph_direction: str = "LR",
    ):
        """
        Print mermaid diagram of this product.

        Args:
            orientations (str, optional): Orientation of mermaid diagram. See: https://mermaid.js.org/syntax/flowchart.html#direction. Defaults to "LR".
            shape_component (str, optional): Shape of mermaid diagram. Defaults to "odd".
            link_type_str (str, optional): Link type of mermaid diagram. Defaults to "-->".
            subgraph (bool, optional): Subgraph or not. Defaults to True.
            subgraph_direction (str, optional): Direction of subgraph. Defaults to "LR".
        """
        self.print_target_component_mermaid_diagram(
            target_component_set=self.component_set,
            orientations=orientations,
            shape_component=shape_component,
            link_type_str=link_type_str,
            subgraph=subgraph,
            subgraph_direction=subgraph_direction,
        )

    def get_gantt_mermaid(
        self,
        target_id_order_list: list[str] = None,
        section: bool = True,
        range_time: tuple[int, int] = (0, sys.maxsize),
        view_ready: bool = False,
        detailed_info: bool = False,
        id_name_dict: dict[str, str] = None,
    ):
        """
        Get mermaid diagram of Gantt chart.

        Args:
            target_id_order_list (list[str], optional): Target ID order list. Defaults to None.
            section (bool, optional): Section or not. Defaults to True.
            range_time (tuple[int, int], optional): Range of Gantt chart. Defaults to (0, sys.maxsize).
            view_ready (bool, optional): If True, ready tasks are included in gantt chart. Defaults to False.
            detailed_info (bool, optional): If True, detailed information is included in gantt chart. Defaults to False.
            id_name_dict (dict[str, str], optional): Dictionary of ID and name for detailed information. Defaults to None.

        Returns:
            list[str]: List of lines for mermaid diagram.
        """
        target_instance_list = self.component_set
        if target_id_order_list is not None:
            id_to_instance = {instance.ID: instance for instance in self.component_set}
            target_instance_list = [
                id_to_instance[tid]
                for tid in target_id_order_list
                if tid in id_to_instance
            ]

        list_of_lines = []
        if section:
            list_of_lines.append(f"section {self.name}")
        for component in target_instance_list:
            list_of_lines.extend(
                component.get_gantt_mermaid_data(
                    range_time=range_time,
                    view_ready=view_ready,
                    detailed_info=detailed_info,
                    id_name_dict=id_name_dict,
                )
            )
        return list_of_lines

    def print_gantt_mermaid(
        self,
        target_id_order_list: list[str] = None,
        date_format: str = "X",
        axis_format: str = "%s",
        section: bool = True,
        range_time: tuple[int, int] = (0, sys.maxsize),
        view_ready: bool = False,
        detailed_info: bool = False,
        id_name_dict: dict[str, str] = None,
    ):
        """
        Print mermaid diagram of Gantt chart.

        Args:
            target_id_order_list (list[str], optional): Target ID order list. Defaults to None.
            date_format (str, optional): Date format of mermaid diagram. Defaults to "X".
            axis_format (str, optional): Axis format of mermaid diagram. Defaults to "%s".
            section (bool, optional): Section or not. Defaults to True.
            range_time (tuple[int, int], optional): Range of Gantt chart. Defaults to (0, sys.maxsize).
            view_ready (bool, optional): If True, ready tasks are included in gantt chart. Defaults to False.
            detailed_info (bool, optional): If True, detailed information is included in gantt chart. Defaults to False.
            id_name_dict (dict[str, str], optional): Dictionary of ID and name for detailed information. Defaults to None.
        """
        print("gantt")
        print(f"dateFormat {date_format}")
        print(f"axisFormat {axis_format}")
        list_of_lines = self.get_gantt_mermaid(
            target_id_order_list=target_id_order_list,
            section=section,
            range_time=range_time,
            view_ready=view_ready,
            detailed_info=detailed_info,
            id_name_dict=id_name_dict,
        )
        print(*list_of_lines, sep="\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_task."""

from __future__ import annotations
import abc
import sys
import uuid
from enum import IntEnum
import warnings

from typing import TYPE_CHECKING

from .base_priority_rule import ResourcePriorityRuleMode, WorkplacePriorityRuleMode

if TYPE_CHECKING:
    from pDESy.model.base_component import BaseComponent


class BaseTaskState(IntEnum):
    """BaseTaskState.

    Enum for representing the state of a task.

    Attributes:
        NONE (int): No state.
        READY (int): Ready to start.
        WORKING (int): Currently working.
        WORKING_ADDITIONALLY (int): Working additionally.
        FINISHED (int): Finished.
    """

    NONE = 0
    READY = 1
    WORKING = 2
    WORKING_ADDITIONALLY = 3
    FINISHED = -1


class BaseTaskDependency(IntEnum):
    """BaseTaskDependency.

    Enum for representing task dependency types.

    Attributes:
        FS (int): Finish to Start.
        SS (int): Start to Start.
        FF (int): Finish to Finish.
        SF (int): Start to Finish.
    """

    FS = 0  # Finish to Start
    SS = 1  # Start to Start
    FF = 2  # Finish to Finish
    SF = 3  # Finish to Start


class BaseTask(object, metaclass=abc.ABCMeta):
    """BaseTask.

    BaseTask class for expressing target workflow. This class will be used as a template.

    Args:
        name (str, optional): Name of this task. Defaults to None -> "New Task".
        ID (str, optional): ID will be defined automatically. Defaults to None -> str(uuid.uuid4()).
        default_work_amount (float, optional): Default workamount of this BaseTask. Defaults to None -> 10.0.
        work_amount_progress_of_unit_step_time (float, optional): Baseline of work amount progress of unit step time. Default to None -> 1.0.
        input_task_id_dependency_set (set(tuple(str, BaseTaskDependency)), optional): Set of input BaseTask id and type of dependency(FS, SS, SF, F/F) tuple. Defaults to None -> set().
        allocated_team_id_set (set[str], optional): Set of allocated BaseTeam id. Defaults to None -> set().
        allocated_workplace_id_set (set[str], optional): Set of allocated BaseWorkplace id. Defaults to None -> set().
        parent_workflow_id (str, optional): Parent workflow id. Defaults to None.
        workplace_priority_rule (WorkplacePriorityRuleMode, optional): Workplace priority rule for simulation. Defaults to WorkplacePriorityRuleMode.FSS.
        worker_priority_rule (ResourcePriorityRule, optional): Worker priority rule for simulation. Defaults to ResourcePriorityRule.SSP.
        facility_priority_rule (ResourcePriorityRule, optional): Task priority rule for simulation. Defaults to TaskPriorityRule.TSLACK.
        need_facility (bool, optional): Whether one facility is needed for performing this task or not. Defaults to False.
        target_component_id (str, optional): Target BaseComponent id. Defaults to None.
        default_progress (float, optional): Progress before starting simulation (0.0 ~ 1.0). Defaults to None -> 0.0.
        due_time (int, optional): Due time. Defaults to None -> int(-1).
        auto_task (bool, optional): If True, this task is performed automatically even if there are no allocated workers. Defaults to False.
        fixing_allocating_worker_id_set (set[str], optional): Allocating worker ID set for fixing allocation in simulation. Defaults to None.
        fixing_allocating_facility_id_set (set[str], optional): Allocating facility ID set for fixing allocation in simulation. Defaults to None.
        est (float, optional): Earliest start time of CPM. This will be updated step by step. Defaults to 0.0.
        eft (float, optional): Earliest finish time of CPM. This will be updated step by step. Defaults to 0.0.
        lst (float, optional): Latest start time of CPM. This will be updated step by step. Defaults to -1.0.
        lft (float, optional): Latest finish time of CPM. This will be updated step by step. Defaults to -1.0.
        remaining_work_amount (float, optional): Remaining workamount in simulation. Defaults to None -> default_work_amount * (1.0 - default_progress).
        remaining_work_amount_record_list (List[float], optional): Record of remaining workamount in simulation. Defaults to None -> [].
        state (BaseTaskState, optional): State of this task in simulation. Defaults to BaseTaskState.NONE.
        state_record_list (List[BaseTaskState], optional): Record list of state. Defaults to None -> [].
        allocated_worker_facility_id_tuple_set (set(tuple(str, str)), optional): State of allocating worker and facility id tuple set in simulation. Defaults to None -> set().
        allocated_worker_facility_id_tuple_set_record_list (List[set[tuple(str, str)]], optional): State of allocating worker and facility id tuple set in simulation. Defaults to None -> [].
        additional_work_amount (float, optional): Advanced parameter. Defaults to None.
        additional_task_flag (bool, optional): Advanced variable. Defaults to False.
        actual_work_amount (float, optional): Advanced variable. Default to None -> default_work_amount*(1.0-default_progress)
    """

    def __init__(
        self,
        # Basic parameters
        name: str = None,
        ID: str = None,
        default_work_amount: float = 10.0,
        work_amount_progress_of_unit_step_time: float = 1.0,
        input_task_id_dependency_set: set[tuple[str, BaseTaskDependency]] = None,
        allocated_team_id_set: set[str] = None,
        allocated_workplace_id_set: set[str] = None,
        parent_workflow_id: str = None,
        workplace_priority_rule: WorkplacePriorityRuleMode = WorkplacePriorityRuleMode.FSS,
        worker_priority_rule: ResourcePriorityRuleMode = ResourcePriorityRuleMode.MW,
        facility_priority_rule: ResourcePriorityRuleMode = ResourcePriorityRuleMode.SSP,
        need_facility: bool = False,
        target_component_id: str = None,
        default_progress: float = 0.0,
        due_time: int = int(-1),
        auto_task: bool = False,
        fixing_allocating_worker_id_set: set[str] = None,
        fixing_allocating_facility_id_set: set[str] = None,
        # Basic variables
        est: float = 0.0,
        eft: float = 0.0,
        lst: float = -1.0,
        lft: float = -1.0,
        remaining_work_amount: float = None,
        remaining_work_amount_record_list: list[float] = None,
        state: BaseTaskState = BaseTaskState.NONE,
        state_record_list: list[BaseTaskState] = None,
        allocated_worker_facility_id_tuple_set: set[tuple[str, str]] = None,
        allocated_worker_facility_id_tuple_set_record_list: list[
            set[tuple[str, str]]
        ] = None,
        # Advanced parameters for customized simulation
        additional_work_amount: float = None,
        # Advanced variables for customized simulation
        additional_task_flag: bool = False,
        actual_work_amount: float = None,
    ):
        """init."""
        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name if name is not None else "New Task"
        self.ID = ID if ID is not None else str(uuid.uuid4())
        self.default_work_amount = (
            default_work_amount if default_work_amount is not None else 10.0
        )
        self.work_amount_progress_of_unit_step_time = (
            work_amount_progress_of_unit_step_time
            if work_amount_progress_of_unit_step_time is not None
            else 1.0
        )
        self.input_task_id_dependency_set = (
            input_task_id_dependency_set
            if input_task_id_dependency_set is not None
            else set()
        )
        self.allocated_team_id_set = (
            allocated_team_id_set if allocated_team_id_set is not None else set()
        )
        self.allocated_workplace_id_set = (
            allocated_workplace_id_set
            if allocated_workplace_id_set is not None
            else set()
        )
        self.parent_workflow_id = (
            parent_workflow_id if parent_workflow_id is not None else None
        )
        self.workplace_priority_rule = (
            workplace_priority_rule
            if workplace_priority_rule is not None
            else WorkplacePriorityRuleMode.FSS
        )
        self.worker_priority_rule = (
            worker_priority_rule
            if worker_priority_rule is not None
            else ResourcePriorityRuleMode.SSP
        )
        self.facility_priority_rule = (
            facility_priority_rule
            if facility_priority_rule is not None
            else ResourcePriorityRuleMode.SSP
        )
        self.need_facility = need_facility
        self.target_component_id = (
            target_component_id if target_component_id is not None else None
        )
        self.default_progress = (
            default_progress if default_progress is not None else 0.0
        )
        self.due_time = due_time or int(-1)
        self.auto_task = auto_task if auto_task is not False else False
        self.fixing_allocating_worker_id_set = fixing_allocating_worker_id_set or None
        self.fixing_allocating_facility_id_set = (
            fixing_allocating_facility_id_set or None
        )

        # ----
        # Changeable variable on simulation
        # --
        # Basic variables
        self.est = est if est != 0.0 else 0.0  # Earliest start time
        self.eft = eft if eft != 0.0 else 0.0  # Earliest finish time
        self.lst = lst if lst != 0.0 else -1.0  # Latest start time
        self.lft = lft if lft != 0.0 else -1.0  # Latest finish time
        if remaining_work_amount is not None:
            self.remaining_work_amount = remaining_work_amount
        else:
            self.remaining_work_amount = self.default_work_amount * (
                1.0 - self.default_progress
            )
        self.remaining_work_amount_record_list = remaining_work_amount_record_list or []
        self.state = state or BaseTaskState.NONE
        self.state_record_list = state_record_list or []
        self.allocated_worker_facility_id_tuple_set = (
            frozenset(allocated_worker_facility_id_tuple_set)
            if allocated_worker_facility_id_tuple_set is not None
            else frozenset()
        )
        self.allocated_worker_facility_id_tuple_set_record_list = (
            allocated_worker_facility_id_tuple_set_record_list or []
        )
        # --
        # Advanced parameter for customized simulation
        self.additional_work_amount = additional_work_amount or 0.0
        # --
        # Advanced variables for customized simulation
        self.additional_task_flag = additional_task_flag or False
        self.actual_work_amount = self.default_work_amount * (
            1.0 - self.default_progress
        )

    def __str__(self):
        """Return the name of BaseTask.

        Returns:
            str: Name of BaseTask.
        """
        return f"{self.name}"

    def export_dict_json_data(self):
        """
        Export the information of this task to JSON data.

        Returns:
            dict: JSON format data.
        """
        dict_json_data = {}
        dict_json_data.update(
            type=self.__class__.__name__,
            name=self.name,
            ID=self.ID,
            default_work_amount=float(self.default_work_amount),
            work_amount_progress_of_unit_step_time=self.work_amount_progress_of_unit_step_time,
            input_task_id_dependency_set=[
                [task_id, int(dependency)]
                for (task_id, dependency) in self.input_task_id_dependency_set
            ],
            allocated_team_id_set=list(self.allocated_team_id_set),
            allocated_workplace_id_set=list(self.allocated_workplace_id_set),
            need_facility=self.need_facility,
            target_component_id=(
                self.target_component_id
                if self.target_component_id is not None
                else None
            ),
            default_progress=self.default_progress,
            due_time=self.due_time,
            auto_task=self.auto_task,
            fixing_allocating_worker_id_set=(
                list(self.fixing_allocating_worker_id_set)
                if self.fixing_allocating_worker_id_set is not None
                else None
            ),
            fixing_allocating_facility_id_set=(
                list(self.fixing_allocating_facility_id_set)
                if self.fixing_allocating_facility_id_set is not None
                else None
            ),
            est=float(self.est),
            eft=float(self.eft),
            lst=float(self.lst),
            lft=float(self.lft),
            remaining_work_amount=float(self.remaining_work_amount),
            remaining_work_amount_record_list=[
                float(rwa) for rwa in self.remaining_work_amount_record_list
            ],
            state=int(self.state),
            state_record_list=[int(state) for state in self.state_record_list],
            allocated_worker_facility_id_tuple_set=[
                [worker_id, facility_id]
                for (
                    worker_id,
                    facility_id,
                ) in self.allocated_worker_facility_id_tuple_set
            ],
            allocated_worker_facility_id_tuple_set_record_list=[
                [
                    [worker_id, facility_id]
                    for (
                        worker_id,
                        facility_id,
                    ) in allocated_worker_facility_id_tuple_set
                ]
                for allocated_worker_facility_id_tuple_set in self.allocated_worker_facility_id_tuple_set_record_list
            ],
        )
        return dict_json_data

    def add_target_component(self, target_component: BaseComponent):
        """
        Add target component to this task.

        Args:
            target_component (BaseComponent): Target BaseComponent.
        """
        self.target_component_id = target_component.ID
        target_component.targeted_task_id_set.add(self.ID)

    def append_input_task_dependency(
        self,
        input_task: BaseTask,
        task_dependency_mode: BaseTaskDependency = BaseTaskDependency.FS,
    ):
        """
        Append input task to `input_task_id_dependency_set`.

        .. deprecated:: Use add_input_task instead.

        Args:
            input_task (BaseTask): Input BaseTask.
            task_dependency_mode (BaseTaskDependency, optional): Task Dependency mode between input_task to this task. Default to BaseTaskDependency.FS.
        """
        warnings.warn(
            "append_input_task_dependency is deprecated, use add_input_task instead.",
            DeprecationWarning,
        )
        self.input_task_id_dependency_set.add((input_task.ID, task_dependency_mode))

    def extend_input_task_list(
        self,
        input_task_list: list[BaseTask],
        input_task_dependency_mode: BaseTaskDependency = BaseTaskDependency.FS,
    ):
        """
        Extend the set of input tasks and FS dependency to `input_task_id_dependency_set`.

        .. deprecated:: Use update_input_task_set instead.

        Args:
            input_task_id_dependency_set (set(tuple(str, BaseTask))): Set of input BaseTask and type of dependency(FS, SS, SF, F/F).
        """
        warnings.warn(
            "extend_input_task_list is deprecated, use update_input_task_set instead.",
            DeprecationWarning,
        )
        for input_task in input_task_list:
            self.input_task_id_dependency_set.add(
                (input_task.ID, input_task_dependency_mode)
            )

    def add_input_task(
        self,
        input_task: BaseTask,
        task_dependency_mode: BaseTaskDependency = BaseTaskDependency.FS,
    ):
        """
        Add input task to `input_task_id_dependency_set`.

        Args:
            input_task (BaseTask): Input BaseTask.
            task_dependency_mode (BaseTaskDependency, optional): Task Dependency mode between input_task to this task. Default to BaseTaskDependency.FS.
        """
        self.input_task_id_dependency_set.add((input_task.ID, task_dependency_mode))

    def update_input_task_set(
        self,
        input_task_set: set[BaseTask],
        input_task_dependency_mode: BaseTaskDependency = BaseTaskDependency.FS,
    ):
        """
        Update the set of input tasks and FS dependency to `input_task_id_dependency_set`.

        Args:
            input_task_id_dependency_set (set(tuple(str, BaseTask))): Set of input BaseTask and type of dependency(FS, SS, SF, F/F).
        """
        for input_task in input_task_set:
            self.add_input_task(input_task, input_task_dependency_mode)

    def initialize(
        self, error_tol: float = 1e-10, state_info: bool = True, log_info: bool = True
    ):
        """
        Initialize the changeable variables of BaseTask.

        If `state_info` is True, the following attributes are initialized:
            - `est`
            - `eft`
            - `lst`
            - `lft`
            - `remaining_work_amount`
            - `state`
            - `allocated_worker_facility_id_tuple_set`
            - `additional_task_flag`
            - `actual_work_amount`

        If `log_info` is True, the following attributes are initialized:
            - `remaining_work_amount_record_list`
            - `state_record_list`
            - `allocated_worker_facility_id_tuple_set_record_list`

        Args:
            error_tol (float, optional): Error toleration of work amount for checking the state of this task. Defaults to 1e-10.
            state_info (bool, optional): Whether to initialize state information. Defaults to True.
            log_info (bool, optional): Whether to initialize log information. Defaults to True.
        """
        if state_info:
            self.est = 0.0  # Earliest start time
            self.eft = 0.0  # Earliest finish time
            self.lst = -1.0  # Latest start time
            self.lft = -1.0  # Latest finish time
            self.remaining_work_amount = self.default_work_amount * (
                1.0 - self.default_progress
            )
            self.state = BaseTaskState.NONE
            self.allocated_worker_facility_id_tuple_set = frozenset()
            self.additional_task_flag = False
            self.actual_work_amount = self.default_work_amount * (
                1.0 - self.default_progress
            )
        if log_info:
            self.remaining_work_amount_record_list = []
            self.state_record_list = []
            self.allocated_worker_facility_id_tuple_set_record_list = []

        if state_info and log_info:
            if self.default_progress >= (1.00 - error_tol):
                self.state = BaseTaskState.FINISHED

    def set_alloc_pairs(self, pairs_iterable):
        """
        Set allocated pairs (non-destructive).
        """
        self.allocated_worker_facility_id_tuple_set = frozenset(pairs_iterable)

    def add_alloc_pair(self, pair: tuple[str, str]):
        """
        Add one allocated pair (non-destructive).
        """
        cur = self.allocated_worker_facility_id_tuple_set
        if pair in cur:
            return
        self.allocated_worker_facility_id_tuple_set = frozenset((*cur, pair))

    def remove_alloc_pair(self, pair: tuple[str, str]):
        """Remove one allocated pair (non-destructive)."""
        cur = self.allocated_worker_facility_id_tuple_set
        if pair not in cur:
            return
        self.allocated_worker_facility_id_tuple_set = frozenset(
            x for x in cur if x != pair
        )

    def update_alloc_pairs(self, add=(), remove=()):
        """
        Update allocated pairs (non-destructive).
        """
        cur = self.allocated_worker_facility_id_tuple_set
        if not add and not remove:
            return
        # Create a mutable set from the current frozenset
        s = set(cur)
        if remove:
            s.difference_update(remove)
        if add:
            s.update(add)
        self.allocated_worker_facility_id_tuple_set = frozenset(s)

    def reverse_log_information(self):
        """Reverse log information of all."""
        self.remaining_work_amount_record_list.reverse()
        self.state_record_list.reverse()
        self.allocated_worker_facility_id_tuple_set_record_list.reverse()

    def get_state_from_record(self, time: int):
        """
        Get the state information in time.

        Args:
            time (int): Target simulation time.

        Returns:
            BaseTaskState: Task State information.
        """
        return self.state_record_list[time]

    def remove_absence_time_list(self, absence_time_list: list[int]):
        """
        Remove record information on `absence_time_list`.

        Args:
            absence_time_list (List[int]): List of absence step time in simulation.
        """
        for step_time in sorted(absence_time_list, reverse=True):
            if step_time < len(self.state_record_list):
                self.remaining_work_amount_record_list.pop(step_time)
                self.allocated_worker_facility_id_tuple_set_record_list.pop(step_time)
                self.state_record_list.pop(step_time)

    def insert_absence_time_list(self, absence_time_list: list[int]):
        """
        Insert record information on `absence_time_list`.

        Args:
            absence_time_list (List[int]): List of absence step time in simulation.
        """
        for step_time in sorted(absence_time_list):
            if step_time < len(self.state_record_list):
                if step_time == 0:
                    self.remaining_work_amount_record_list.insert(
                        self.default_work_amount * (1.0 - self.default_progress)
                    )
                    self.allocated_worker_facility_id_tuple_set_record_list.insert(
                        step_time, None
                    )
                    self.state_record_list.insert(step_time, BaseTaskState.NONE)
                else:
                    self.remaining_work_amount_record_list.insert(
                        step_time, self.remaining_work_amount_record_list[step_time - 1]
                    )
                    self.allocated_worker_facility_id_tuple_set_record_list.insert(
                        step_time,
                        self.allocated_worker_facility_id_tuple_set_record_list[
                            step_time - 1
                        ],
                    )

                    insert_state_before = self.state_record_list[step_time - 1]
                    insert_state_after = self.state_record_list[step_time]
                    if insert_state_before == BaseTaskState.WORKING:
                        if insert_state_after == BaseTaskState.FINISHED:
                            insert_state = BaseTaskState.FINISHED
                        else:
                            insert_state = BaseTaskState.READY
                        self.state_record_list.insert(step_time, insert_state)
                    elif (
                        insert_state_before == BaseTaskState.NONE
                        and insert_state_after == BaseTaskState.WORKING
                    ):
                        self.state_record_list.insert(step_time, BaseTaskState.READY)
                    else:
                        self.state_record_list.insert(
                            step_time, self.state_record_list[step_time - 1]
                        )

    def print_log(self, target_step_time: int):
        """
        Print log in `target_step_time`.

        Prints:
            - ID
            - name
            - default_work_amount
            - remaining_work_amount_record_list
            - state_record_list[target_step_time]
            - allocated_worker_facility_id_tuple_set_record_list[target_step_time]

        Args:
            target_step_time (int): Target step time of printing log.
        """
        print(
            self.ID,
            self.name,
            self.default_work_amount,
            max(self.remaining_work_amount_record_list[target_step_time], 0.0),
            self.state_record_list[target_step_time],
            self.allocated_worker_facility_id_tuple_set_record_list[target_step_time],
        )

    def print_all_log_in_chronological_order(self, backward: bool = False):
        """
        Print all log in chronological order.

        Args:
            backward (bool, optional): If True, print logs in reverse order. Defaults to False.
        """
        n = len(self.state_record_list)
        if backward:
            for i in range(n):
                t = n - 1 - i
                print("TIME: ", t)
                self.print_log(t)
        else:
            for t in range(n):
                print("TIME: ", t)
                self.print_log(t)

    def get_time_list_for_gantt_chart(self, finish_margin: float = 1.0):
        """
        Get ready/working time_list for drawing Gantt chart.

        Args:
            finish_margin (float, optional): Margin of finish time in Gantt chart. Defaults to 1.0.

        Returns:
            Tuple[List[tuple(int, int)], List[tuple(int, int)]]:
                - ready_time_list including start_time, length
                - working_time_list including start_time, length
        """
        ready_time_list = []
        working_time_list = []
        previous_state = BaseTaskState.NONE
        from_time = -1
        to_time = -1
        time = -1
        for time, state in enumerate(self.state_record_list):
            if state != previous_state:
                if from_time == -1:
                    from_time = time
                elif to_time == -1:
                    to_time = time
                    if state == BaseTaskState.NONE or state == BaseTaskState.FINISHED:
                        if previous_state == BaseTaskState.WORKING:
                            working_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                        elif previous_state == BaseTaskState.READY:
                            ready_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                    if state == BaseTaskState.READY:
                        if previous_state == BaseTaskState.WORKING:
                            working_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                    if state == BaseTaskState.WORKING:
                        if previous_state == BaseTaskState.READY:
                            ready_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                    from_time = time
                    to_time = -1
            previous_state = state

        # Suspended because of max time limitation
        if from_time > -1 and to_time == -1:
            if previous_state == BaseTaskState.WORKING:
                working_time_list.append((from_time, time - from_time + finish_margin))
            elif previous_state == BaseTaskState.READY:
                ready_time_list.append((from_time, time - from_time + finish_margin))

        # Append dummy values (0, 0) to the lists if they are empty.
        # This ensures that the Gantt chart generation process has valid data to work with,
        # even if no actual time intervals were recorded. Without this, downstream code
        # might encounter errors or render incomplete charts.
        if len(ready_time_list) == 0:
            ready_time_list.append((0, 0))
        if len(working_time_list) == 0:
            working_time_list.append((0, 0))

        return ready_time_list, working_time_list

    def get_mermaid_diagram(
        self,
        shape: str = "rect",
        print_work_amount_info: bool = True,
        subgraph: bool = False,
        subgraph_name: str = "Task",
        subgraph_direction: str = "LR",
    ):
        """
        Get mermaid diagram of this task.

        Args:
            shape (str, optional): Shape of mermaid diagram. Defaults to "rect".
            print_work_amount_info (bool, optional): Print work amount information or not. Defaults to True.
            subgraph (bool, optional): Subgraph or not. Defaults to False.
            subgraph_name (str, optional): Subgraph name. Defaults to "Task".
            subgraph_direction (str, optional): Direction of subgraph. Defaults to "LR".

        Returns:
            list[str]: List of lines for mermaid diagram.
        """
        list_of_lines = []
        if subgraph:
            list_of_lines.append(f"subgraph {subgraph_name}")
            list_of_lines.append(f"direction {subgraph_direction}")
        node_label = self.name
        if print_work_amount_info:
            node_label += f"<br>{self.remaining_work_amount}"
        list_of_lines.append(f"{self.ID}@{{shape: {shape}, label: '{node_label}'}}")

        if subgraph:
            list_of_lines.append("end")

        return list_of_lines

    def print_mermaid_diagram(
        self,
        orientations: str = "LR",
        shape: str = "rect",
        print_work_amount_info: bool = True,
        subgraph: bool = False,
        subgraph_name: str = "Task",
        subgraph_direction: str = "LR",
    ):
        """
        Print mermaid diagram of this task.

        Args:
            orientations (str, optional): Orientation of mermaid diagram. Defaults to "LR".
            shape (str, optional): Shape of mermaid diagram. Defaults to "rect".
            print_work_amount_info (bool, optional): Print work amount information or not. Defaults to True.
            subgraph (bool, optional): Subgraph or not. Defaults to False.
            subgraph_name (str, optional): Subgraph name. Defaults to "Task".
            subgraph_direction (str, optional): Direction of subgraph. Defaults to "LR".
        """
        print(f"flowchart {orientations}")
        list_of_lines = self.get_mermaid_diagram(
            shape=shape,
            print_work_amount_info=print_work_amount_info,
            subgraph=subgraph,
            subgraph_name=subgraph_name,
            subgraph_direction=subgraph_direction,
        )
        print(*list_of_lines, sep="\n")

    def get_gantt_mermaid_data(
        self,
        range_time: tuple[int, int] = (0, sys.maxsize),
        detailed_info: bool = False,
        id_name_dict: dict[str, str] = None,
        view_ready: bool = False,
    ):
        """
        Get gantt mermaid data of this task.

        Args:
            range_time (tuple[int, int], optional): Range of gantt chart. Defaults to (0, sys.maxsize).
            detailed_info (bool, optional): If True, print detailed information. Defaults to False.
            id_name_dict (dict[str, str], optional): Dictionary to map ID to name. Defaults to None.
            view_ready (bool, optional): If True, include ready time in Gantt chart. Defaults to False.

        Returns:
            list[str]: List of lines for gantt mermaid diagram.
        """
        list_of_lines = []
        ready_time_list, working_time_list = self.get_time_list_for_gantt_chart()
        if view_ready:
            for start, duration in ready_time_list:
                end = start + duration - 1
                if end < range_time[0] or start > range_time[1]:
                    continue
                clipped_start = max(start, range_time[0])
                clipped_end = min(end + 1, range_time[1])

                text = self.name + "[READY]"
                if (
                    detailed_info is True
                    and id_name_dict is not None
                    and self.ID in id_name_dict
                    and clipped_start
                    < len(self.allocated_worker_facility_id_tuple_set_record_list)
                ):
                    worker_facility_tuple_list = (
                        self.allocated_worker_facility_id_tuple_set_record_list[
                            clipped_start
                        ]
                    )
                    if worker_facility_tuple_list is None:
                        worker_facility_tuple_list = ()
                    worker_id_list = [
                        worker_id
                        for worker_id, _ in worker_facility_tuple_list
                        if worker_id is not None
                    ]
                    worker_name_list = [
                        id_name_dict.get(worker_id, worker_id)
                        for worker_id in worker_id_list
                    ]
                    facility_id_list = [
                        facility_id
                        for _, facility_id in worker_facility_tuple_list
                        if facility_id is not None
                    ]
                    facility_name_list = [
                        id_name_dict.get(facility_id, facility_id)
                        for facility_id in facility_id_list
                    ]

                    combined_name_list = []
                    max_length = max(len(worker_name_list), len(facility_name_list))
                    for i in range(max_length):
                        worker_name = (
                            worker_name_list[i] if i < len(worker_name_list) else ""
                        )
                        facility_name = (
                            facility_name_list[i] if i < len(facility_name_list) else ""
                        )

                        if worker_name and facility_name:
                            combined_name_list.append(f"{worker_name}-{facility_name}")
                        elif worker_name:
                            combined_name_list.append(worker_name)
                        elif facility_name:
                            combined_name_list.append(facility_name)

                    if combined_name_list:
                        text = f"{self.name} * {'&'.join(combined_name_list)} [READY]"

                list_of_lines.append(f"{text}:{int(clipped_start)},{int(clipped_end)}")

        for start, duration in working_time_list:
            end = start + duration - 1
            if end < range_time[0] or start > range_time[1]:
                continue
            clipped_start = max(start, range_time[0])
            clipped_end = min(end + 1, range_time[1])

            text = self.name
            if (
                detailed_info is True
                and id_name_dict is not None
                and self.ID in id_name_dict
                and clipped_start
                < len(self.allocated_worker_facility_id_tuple_set_record_list)
            ):
                worker_facility_tuple_list = (
                    self.allocated_worker_facility_id_tuple_set_record_list[
                        clipped_start
                    ]
                )
                if worker_facility_tuple_list is None:
                    worker_facility_tuple_list = ()
                worker_id_list = [
                    worker_id
                    for worker_id, _ in worker_facility_tuple_list
                    if worker_id is not None
                ]
                worker_name_list = [
                    id_name_dict.get(worker_id, worker_id)
                    for worker_id in worker_id_list
                ]
                facility_id_list = [
                    facility_id
                    for _, facility_id in worker_facility_tuple_list
                    if facility_id is not None
                ]
                facility_name_list = [
                    id_name_dict.get(facility_id, facility_id)
                    for facility_id in facility_id_list
                ]

                combined_name_list = []
                max_length = max(len(worker_name_list), len(facility_name_list))
                for i in range(max_length):
                    worker_name = (
                        worker_name_list[i] if i < len(worker_name_list) else ""
                    )
                    facility_name = (
                        facility_name_list[i] if i < len(facility_name_list) else ""
                    )

                    if worker_name and facility_name:
                        combined_name_list.append(f"{worker_name}-{facility_name}")
                    elif worker_name:
                        combined_name_list.append(worker_name)
                    elif facility_name:
                        combined_name_list.append(facility_name)

                if combined_name_list:
                    text = f"{self.name} * {'&'.join(combined_name_list)}"

            list_of_lines.append(f"{text}:{int(clipped_start)},{int(clipped_end)}")
        return list_of_lines
